import items.*;
import pokemon.Pokemon;
import storage.Inventory;
import storage.Trainer;

import java.util.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.lang.NumberFormatException;
import java.text.NumberFormat;
import java.text.ParseException;

/** Simulates the game Pokémon */
public class Game {
    public static final Scanner in = new Scanner(System.in);
    private static Pokemon opponentPokemon;
    private static List<Pokemon> availableOpponentPokemon;
    private static Pokemon playersCurrentPokemon;
    private static Trainer player;

    public static void main(String[] args) {
        availableOpponentPokemon = readPokemon();
        int initialPokemon = availableOpponentPokemon.size();

        System.out.println("What is your name?");
        String name = in.nextLine();
        player = new Trainer(name, randomTeam(), createInventory());

        boolean done = false;

        triggerNewEncounter();

        while (!done) {
            if (opponentPokemon == null) {
                System.out.printf("You have brutally murdered %d pokemon.%n"
                                + "The only ones left are the ones in your possession.%n"
                                + "There really is nothing more for you to do here.%n", initialPokemon);
                return;
            }
            if (player.getConsciousPokemon().size() < 1) {
                System.out.println("All your pokemon have fainted. Your journey ends here.");
                return;
            }
            while (playersCurrentPokemon == null || !playersCurrentPokemon.isConscious()) {
                player.listAvailablePokemon(true);
                playersCurrentPokemon = player.choosePokemon(true);
            }
            System.out.printf("Opponent: %s%nWhat will you do?%n", opponentPokemon);
            System.out.printf("b: battle"
                          + "%nh: heal or revive"
                          + "%nt: throw poke ball"
                          + "%nc: change pokemon"
                          + "%nf: flee"
                          + "%nv: view my pokemon"
                          + "%ns: save"
                          + "%nl: load"
                          + "%nq: quit%n>");
            char command = in.next().toLowerCase().charAt(0);
            switch (command) {
                case 'b':
                    battle();
                    break;
                case 'h':
                    healPokemon();
                    break;
                case 't':
                    throwPokeBall();
                    break;
                case 'c':
                    choosePokemon();
                    break;
                case 's':
                    save();
                    break;
                case 'l':
                    load();
                    break;
                case 'f':
                    flee();
                    break;
                case 'v':
                    showPokemon();
                    break;
                case 'q':
                    done = true;
                    break;
                default:
                    System.out.println("[Invalid command]");
            }
        }
    }

    /**
     * Lists all of the player's pokemon
     */
    private static void showPokemon() {
        player.printPokemon();
    }


    /**
     * Tries to flee from the current opponent
     */
    private static void flee() {
        boolean fleeSuccessful = playersCurrentPokemon.flee(opponentPokemon);
        if (fleeSuccessful) {
            System.out.println("You fled the battle.");
            triggerNewEncounter();
        } else {
            System.out.printf("Failed to flee from %s.%n", opponentPokemon.getName());
            opponentPokemon.attack(playersCurrentPokemon);
        }
    }

    /**
     * Loads game information from a save
     */
    private static void load() {
        List<Pokemon> loadedPokemon = loadPokemon("pokemon.save");
        List<Pokemon> loadedUsersPokemon = loadPokemon("user.save");
        Inventory loadedInventory = loadInventory("inventory.save");
        if (loadedPokemon == null || loadedUsersPokemon == null || loadedInventory == null) {
            System.out.println("One or more save files seem corrupt or missing. Please delete, create or fix the affected file(s).");
        } else {
            availableOpponentPokemon = loadedPokemon;
            player.setPokemon(loadedUsersPokemon);
            player.setInventory(loadedInventory);
            if (availableOpponentPokemon.size() > 0 && player.getPokemon().size() > 0) {
                do {
                    player.listAvailablePokemon(true);
                    playersCurrentPokemon = player.choosePokemon(true);
                } while (playersCurrentPokemon == null || !playersCurrentPokemon.isConscious());
                opponentPokemon = randomPokemon(availableOpponentPokemon);
            }
        }
    }

    /**
     * Saves all stats about the game
     */
    private static void save() {
        savePokemon(availableOpponentPokemon, "pokemon.save");
        savePokemon(player.getPokemon(), "user.save");
        saveInventory(player.getInventory(), "inventory.save");
    }

    /**
     * Chooses a new pokemon from the player's available pokemon
     */
    private static void choosePokemon() {
        player.listAvailablePokemon(true);
        playersCurrentPokemon = player.choosePokemon(true);
        while (playersCurrentPokemon == null) {
            player.listAvailablePokemon(true);
            playersCurrentPokemon = player.choosePokemon(true);
        }
        opponentPokemon.attack(playersCurrentPokemon);
    }

    /**
     * Makes the player's pokemon attack the enemy pokemon
     */
    private static void battle() {
        if (opponentPokemon.isConscious() && playersCurrentPokemon.isConscious()) {
            playersCurrentPokemon.attack(opponentPokemon);
            if (opponentPokemon.isConscious()) {
                opponentPokemon.attack(playersCurrentPokemon);
                if (!playersCurrentPokemon.isConscious()) {
                    System.out.println("Your pokemon fainted.");
                }
            } else {
                pokemonFainted(availableOpponentPokemon, opponentPokemon);
                System.out.println("The opponent pokemon fainted.");
                triggerNewEncounter();
            }
        }
    }

    /**
     * Heals a pokemon if possible
     */
    private static void healPokemon() {
        if (player.getInventory().getPotions().size() > 0) {
            player.getInventory().availablePotions();
            AbstractPotion currentPotion = player.getInventory().getChosenPotion();
            if (currentPotion == null) {
                in.nextLine();
                System.out.println("Invalid potion.");
            } else {
                Pokemon pokemonToHeal;
                if (currentPotion.needsAlive()) {
                    player.listAvailablePokemon(true);
                    pokemonToHeal = player.choosePokemon(true);
                } else {
                    player.listAvailablePokemon(false);
                    pokemonToHeal = player.choosePokemon(false);
                }
                if (pokemonToHeal == null) {
                    System.out.println("That is not a valid pokemon");
                } else {
                    if (currentPotion.use(pokemonToHeal)) {
                        player.getInventory().addItem(currentPotion, -1);
                        opponentPokemon.attack(playersCurrentPokemon);
                    }
                }
            }
        } else {
            System.out.println("You have used all your healing items.");
        }
    }

    /**
     * Tries throwing a poke ball at the enemy pokemon
     */
    private static void throwPokeBall() {
        if (player.getInventory().getPokeBalls().size() > 0) {
            player.getInventory().availablePokeBalls();
            AbstractPokeBall currentPokeBall = player.getInventory().getChosenPokeBall();
            if (currentPokeBall == null) {
                in.nextLine();
                System.out.println("Invalid poke ball.");
            } else {
                boolean captured = currentPokeBall.use(opponentPokemon, availableOpponentPokemon, player);
                player.getInventory().addItem(currentPokeBall, -1);
                if (captured) {
                    triggerNewEncounter();
                } else {
                    opponentPokemon.attack(playersCurrentPokemon);
                }
            }
        } else {
            System.out.println("You have used all your poke balls.");
        }
    }

    /**
     * Triggers an encounter with a new wild pokemon
     */
    private static void triggerNewEncounter() {
        opponentPokemon = randomPokemon(availableOpponentPokemon);
        System.out.printf("A wild %s appeared.%n", opponentPokemon.getName());
    }

    /**
     * Removes a fainted pokemon from encounter-able pokemon
     * @param pokemonList <p>List of pokemon to search</p>
     * @param target <p>The pokemon to remove</p>
     */
    public static void pokemonFainted(List<Pokemon> pokemonList, Pokemon target) {
        pokemonList.removeIf(pokemon -> pokemon.equals(target));
    }

    /**
     * Gives a trainer necessary starting items
     * @return <p>A list of items</p>
     */
    public static Inventory createInventory() {
        Inventory inventory = new Inventory();
        inventory.addItem(new PokeBall(), 15);
        inventory.addItem(new GreatBall(), 10);
        inventory.addItem(new UltraBall(), 5);
        inventory.addItem(new MasterBall(), 1);
        inventory.addItem(new Potion(), 20);
        inventory.addItem(new SuperPotion(), 10);
        inventory.addItem(new HyperPotion(), 5);
        inventory.addItem(new MaxPotion(), 1);
        inventory.addItem(new Revive(), 2);
        return inventory;
    }

    /**
     * Chooses a random pokemon from a list
     * @param pokemon <p>The list to choose from</p>
     * @return <p>A pokemon.Pokemon object, or null if the list is empty</p>
     */
    public static Pokemon randomPokemon(List<Pokemon> pokemon) {
        if (pokemon.size() > 0) {
            return pokemon.get((int)(Math.random() * pokemon.size()));
        } else {
            throw new IllegalArgumentException("Can't get random pokemon from empty list");
        }
    }

    /**
     * Reads pokemon from pokemon.Pokemon.txt to an ArrayList.
     * @return <p>An ArrayList of pokemon objects</p>
     */
    public static ArrayList<Pokemon> readPokemon() {
        ArrayList<Pokemon> pokemon = new ArrayList<>();
        try (Scanner file = new Scanner(new File("config/Pokemon.txt"))) {
            while (file.hasNextLine()) {
                pokemon.add(new Pokemon(file.nextLine()));
            }
        } catch (FileNotFoundException e) {
            /* If the file is compiled as jar, this will prevent an empty list. */
            try (Scanner file = new Scanner(Game.class.getResourceAsStream("config/Pokemon.txt"))) {
                while (file.hasNextLine()) {
                    pokemon.add(new Pokemon(file.nextLine()));
                }
            }
        }
        return pokemon;
    }

    /**
     * Picks a random pokemon from a list
     * @param pokemon <p>A list of pokemon objects</p>
     * @return <p>A pokemon object which exists in the input list</p>
     */
    public static Pokemon pickRandomPokemon(ArrayList<Pokemon> pokemon) {
        int index = (int)(Math.random() * (pokemon.size()));
        Pokemon randomPokemon = pokemon.get(index);
        pokemon.remove(index);
        return randomPokemon;
    }

    /**
     * Saves all pokemon in a list to a text file
     * @param pokemonList <p>List of pokemon.Pokemon objects to save</p>
     * @param saveFile <p>The file to write to</p>
     */
    public static void savePokemon(List<Pokemon> pokemonList, String saveFile) {
        try (PrintWriter file = new PrintWriter(saveFile)) {
            for (Pokemon pokemon : pokemonList) {
                file.println(pokemon.saveString());
            }
            System.out.println("Successfully saved pokemon.");
        } catch (FileNotFoundException e) {
            System.out.println("The saveFile could not be written.");
        }
    }

    /**
     * Saves the inventory to a text file
     * @param inventory <p>The inventory to save</p>
     * @param saveFile <p>The file to write to</p>
     */
    public static void saveInventory(Inventory inventory, String saveFile) {
        try (PrintWriter file = new PrintWriter(saveFile)) {
            file.println(inventory.getSaveString());
            System.out.println("Successfully saved inventory.");
        } catch (FileNotFoundException e) {
            System.out.println("The saveFile could not be written.");
        }
    }

    /**
     * Loads pokemon from a text file
     * @param saveFile <p>The file to read</p>
     * @return <p>A list of pokemon or null on failure</p>
     */
    public static ArrayList<Pokemon> loadPokemon(String saveFile) {
        ArrayList<Pokemon> pokemon = new ArrayList<>();
        try (Scanner file = new Scanner(new File(saveFile))) {
            NumberFormat format = NumberFormat.getInstance(Locale.ENGLISH);
            while (file.hasNextLine()) {
                String[] data = file.nextLine().split(";");
                try {
                    pokemon.add(new Pokemon(data[0], Integer.parseInt(data[1]), Integer.parseInt(data[2]),
                            Integer.parseInt(data[3]), format.parse(data[4]).doubleValue(),
                            Integer.parseInt(data[5]), Integer.parseInt(data[6]), Integer.parseInt(data[7])));
                } catch (NumberFormatException | ParseException e) {
                    System.out.println("Malformed number " + e.getMessage());
                } catch (ArrayIndexOutOfBoundsException e) {
                    System.out.println("Invalid saveFile: " + saveFile);
                    return null;
                }
            }
            System.out.println("Successfully loaded pokemon.");
        } catch (FileNotFoundException e) {
            System.out.println("Missing saveFile " + saveFile);
            return null;
        }
        return pokemon;
    }

    /**
     * Loads inventory from a text file
     * @param saveFile <p>The file to write to</p>
     * @return <p>A list of items or null on failure</p>
     */
    public static Inventory loadInventory(String saveFile) {
        Inventory inventory = new Inventory();
        try (Scanner file = new Scanner(new File(saveFile))) {
            while (file.hasNextLine()) {
                try {
                    String line = file.nextLine();
                    if (line.isEmpty()) {
                        break;
                    }
                    String[] data = line.split(";");
                    inventory.addItem(Item.getItemByName(data[0]), Integer.parseInt(data[1]));
                } catch (ArrayIndexOutOfBoundsException e) {
                    System.out.println("Invalid saveFile: " + saveFile);
                    return null;
                }
            }
            System.out.println("Successfully loaded items.");
        } catch (FileNotFoundException e) {
            System.out.println("Missing saveFile " + saveFile);
            return null;
        }
        return inventory;
    }

    /**
     * Returns a random list of 6 pokemon
     * @return <p>A list of pokemon.Pokemon</p>
     */
    public static ArrayList<Pokemon> randomTeam() {
        ArrayList<Pokemon> temporary = readPokemon();
        ArrayList<Pokemon> pokemon = new ArrayList<>();
        for (int i = 1; i <= 6; i++) {
            pokemon.add(pickRandomPokemon(temporary));
        }
        return pokemon;
    }
}
