package pokemon;

import storage.Trainer;

import java.util.List;
import java.util.Random;

/**
 * This class represents a pokemon
 */
public class Pokemon {
    private final String name;
    private int healthPoints;
    private int maxHealthPoints;
    private int strength;
    private final double criticalChance;
    private final Random random;
    private final int catchRate;
    private int exp;
    private int level;
    private int fleeCount;

    /**
     * Instantiates a new pokemon with random stats
     * @param name <p>The name of the pokemon</p>
     */
    public Pokemon(String name) {
        this.random = new Random();
        this.name = name;
        this.healthPoints = 30 + random.nextInt(70);
        this.maxHealthPoints = this.healthPoints;
        this.criticalChance = Math.abs(0.1 * random.nextGaussian());
        this.catchRate = (int)(Math.random() * 256);
        this.exp = 0;
        this.level = (int)(Math.abs(2 * random.nextGaussian()) + 1);
        this.strength = random.nextInt(7) + this.level;
        this.fleeCount = 0;
    }

    /**
     * Instantiates a pokemon with set stats
     * @param name <p>The name of the pokemon</p>
     * @param healthPoints <p>The current HP of the pokemon</p>
     * @param maxHealthPoints <p>The max HP of the pokemon</p>
     * @param strength <p>The strength of the pokemon</p>
     * @param criticalChance <p>The critical hit chance of the pokemon</p>
     * @param catchRate <p>The catch rate of the pokemon. A higher rate makes it easier to catch</p>
     * @param exp <p>The current xp of the pokemon</p>
     * @param level <p>The current level of the pokemon</p>
     */
    public Pokemon(String name, int healthPoints, int maxHealthPoints, int strength, double criticalChance, int catchRate, int exp, int level) {
        this.random = new Random();
        this.name = name;
        this.healthPoints = healthPoints;
        this.maxHealthPoints = maxHealthPoints;
        this.criticalChance = criticalChance;
        this.catchRate = catchRate;
        this.exp = exp;
        this.level = level;
        this.strength = strength;
        this.fleeCount = 0;
    }

    /**
     * Makes a string containing all information about the pokemon
     * @return <p>A descriptive string</p>
     */
    public String toString() {
        return String.format("Level %d %s HP: (%d/%d) STR: %d CHC: %.0f%%", this.level, this.name, this.healthPoints,
                this.maxHealthPoints, this.strength, (this.criticalChance*100));
    }

    /**
     * Returns a string containing all the pokemon's data in a save-able format
     */
    public String saveString() {
        return String.format("%s;%d;%d;%d;%f;%d;%d;%d", this.name, this.healthPoints, this.maxHealthPoints,
                this.strength, this.criticalChance, this.catchRate, this.exp, this.level);
    }

    /**
     * Gets the name of the pokemon
     * @return <p>The name of the pokemon</p>
     */
    public String getName() {
        return this.name;
    }

    /**
     * Gets the current hit points of the pokemon
     * @return <p>The current HP of the pokemon</p>
     */
    public int getHP() {
        return this.healthPoints;
    }

    /**
     * Checks whether two pokemon are equal
     * @param target <p>The pokemon to check against</p>
     * @return <p>True if they are equal. False otherwise</p>
     */
    public boolean equals(Pokemon target) {
        return target == this;
    }

    /**
     * Heals a pokemon
     * @param amount <p>How many hit points to heal</p>
     * @return <p>The exact amount of hit points healed</p>
     */
    public int heal(int amount) {
        if (amount < 0) {
            amount = this.maxHealthPoints - this.healthPoints;
        } else {
            amount = Math.min(this.maxHealthPoints - this.healthPoints, amount);
        }
        this.healthPoints += amount;
        return amount;
    }

    /**
     * Checks if a pokemon has not taken any damage or is fully healed
     * @return <p>True if health is full. False otherwise</p>
     */
    public boolean isDamaged() {
        return this.healthPoints < this.maxHealthPoints;
    }

    /**
     * Tries to catch a pokemon
     * @param current <p>The list where the pokemon currently belongs</p>
     * @param catcher <p>The list to send the pokemon on successful capture</p>
     * @return <p>True on successful capture. False otherwise</p>
     */
    public boolean tryCapture(List<Pokemon> current, Trainer catcher, int range) {
        if (range == 0) {
            this.capture(current, catcher);
            System.out.printf("%s was caught.%n", this.name);
            return true;
        }
        if ((int)(Math.random() * (range + 1)) < Math.max((3 * this.maxHealthPoints - 2 * this.healthPoints) *
                this.catchRate / (3 * this.maxHealthPoints), 1)) {
            this.capture(current, catcher);
            System.out.printf("%s was caught.%n", this.name);
            return true;
        } else {
            System.out.printf("%s escaped from the poke ball.%n", this.name);
            return false;
        }
    }

    /**
     * Captures a wild pokemon
     * @param current <p>The pokemon list the pokemon belongs to</p>
     * @param trainer <p>The trainer capturing the pokemon</p>
     */
    private void capture(List<Pokemon> current, Trainer trainer) {
        trainer.addPokemon(this);
        current.removeIf(pokemon -> pokemon == this);
    }

    /**
     * Restores the health of a fainted pokemon
     */
    public void revive() {
        this.healthPoints = this.maxHealthPoints;
    }

    /**
     * Checks if the pokemon has any HP left
     * @return <p>True if any HP is left. False otherwise</p>
     */
    public boolean isConscious() {
        return this.healthPoints > 0;
    }

    /**
     * Damages a pokemon
     * @param damageTaken <p>How many hit points are to be deducted</p>
     */
    public void damage(int damageTaken) {
        this.healthPoints = Math.max(this.healthPoints -= damageTaken, 0);
        System.out.printf("%s takes %d damage and is left with %d/%d HP%n", this.name, damageTaken, this.healthPoints, this.maxHealthPoints);
    }

    /**
     * Gives a pokemon exp after each successful battle. Also handles leveling up
     * @param target <p>Which pokemon did we beat</p>
     */
    private void giveEXP(Pokemon target) {
        int exp = (100 * target.level)/7;
        this.exp += exp;
        System.out.printf("%s gained %d exp.%n", this.name, exp);
        if (this.exp > (4 * Math.pow(this.level, 3)) / 5) {
            this.level++;
            this.maxHealthPoints += 5;
            this.healthPoints = maxHealthPoints;
            this.strength += 1;
            System.out.printf("%s is now level %d. Health has been increased and restored. Strength has been increased.%n", this.name, this.level);
        }
    }

    /**
     * Makes a pokemon attack another pokemon
     * @param target <p>The pokemon to take damage</p>
     */
    public void attack(Pokemon target) {
        if (this.healthPoints > 0) {
            System.out.printf("%s attacks %s.%n", this.getName(), target.getName());
            double critical = 1;
            if (Math.random() < this.criticalChance) {
                System.out.println("Critical hit!");
                critical = (2.0 * this.level +5) / (this.level + 5);
            }
            double randomDouble = 0.85 + (1.0 - 0.85) * random.nextDouble();
            int moveDamage = (int)(Math.random() * 60 + 40);
            int damageInflicted = this.level + (int)((((((2 * this.strength) / 5) + 2) * moveDamage * 1.75) / 50 + 2) * critical * randomDouble);
            target.damage(damageInflicted);
            if (!target.isConscious()) {
                System.out.printf("%s is defeated by %s.%n", target.getName(), this.getName());
                this.giveEXP(target);
                this.fleeCount = 0;
            }
        } else {
            System.out.println("No cheating!");
        }
    }

    /**
     * The trainer may flee if the battle is too hard
     * @param target <p>Who are we trying to flee from</p>
     * @return <p>True on successful escape. False otherwise</p>
     */
    public boolean flee(Pokemon target) {
        if (random.nextInt(256) < (this.level * 128 / target.level) + (30 * this.fleeCount % 256)) {
            this.fleeCount = 0;
            return true;
        } else {
            this.fleeCount++;
            return false;
        }
    }
}
